#include "network.h"
#include "sigint.h"
#include "main.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <error.h>
#include <errno.h>
#include <sys/errno.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>

network::network()
{
}
 
network::network(char *host, int p)
{
  struct hostent *he;
  struct sockaddr_in peer;
  int on = 1, status, conStatus = 0;
  
  // Clear some variables
  len = 0;
  buf = new char [BUF_SIZE];
  fd = 1;
  if (fd < 0) throw couldNotOpen();
  localSecure = 0;
  remoteSecure = 0;
  encryption = -1;
  handshook = false;
  localIV = 0;
  remoteIV = 0;
  master = false;
  localCryptor = 0;
  remoteCryptor = 0;
  closeHandle = false;
  
  if(handshook != false)
    {
      printf("network::network(char *, int): handshook -:%d:-\n", handshook);
      handshook = false;
    }
  
  if (verbose) 
    printf("NET: Making connection to %s:%i.\n", host, p);
  
  he = gethostbyname(host);
  if (!he)
    {
      printf("gethostbyname failed\n");
      he = gethostbyaddr(host, sizeof(host), AF_INET);
      if (!he) throw couldNotOpen();
    }
  
  do
    {
      if( conStatus < 0 )
	{
	  close( fd );
	  sleep(60);
	}
      
      fd = socket(PF_INET, SOCK_STREAM, 0);
      if (fd < 0)
	{
	  printf("socket failed\n");
	  throw couldNotOpen();
	}
      
      // Enable reuse of the same socket
      status = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const char *) &on, sizeof(on));
      if (status < 0 ) throw couldNotOpen();
      
      // Disable Naegle's algorithm
      status = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (const char *) &on, sizeof(on));
      if (status < 0 ) throw couldNotOpen();
      
      peer.sin_family = AF_INET;
      peer.sin_port = htons(p);
      peer.sin_addr = *(struct in_addr*) he->h_addr;
    }
  while(( conStatus = connect(fd, (struct sockaddr*) &peer, sizeof(peer))) < 0);
  
  if (verbose) 
    printf("NET: Connected.\n");
  
  closeHandle = true;
}

network::network(int p)
{
  struct sockaddr_in sa, peer;
  fd_set fs;
  int fd_listen, status, on=1;
  unsigned int l;
  
  // Clear some variables
  len = 0;
  buf = new char [BUF_SIZE];
  fd = 1;
  if (fd < 0) throw couldNotOpen();
  localSecure = 0;
  remoteSecure = 0;
  encryption = -1;
  handshook = false;
  localIV = 0;
  remoteIV = 0;
  master = false;
  localCryptor = 0;
  remoteCryptor = 0;
  closeHandle = false;
  
  if (verbose) 
    printf("NET: Waiting for connection on port %i.\n", p);
  
  fd_listen = socket(PF_INET, SOCK_STREAM, 0);
  if (fd_listen < 0) throw couldNotOpen();
  
  // Enable reuse of the same socket
  status = setsockopt(fd_listen, SOL_SOCKET, SO_REUSEADDR, (const char *) &on, sizeof(on));
  if (status < 0 ) throw couldNotOpen();
  
  // Disable Naegle's algorithm
  status = setsockopt(fd_listen, IPPROTO_TCP, TCP_NODELAY, (const char *) &on, sizeof(on));
  if (status < 0 ) throw couldNotOpen();
  
  sa.sin_family = AF_INET;
  sa.sin_port = htons(p);
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  
  if (bind(fd_listen, (struct sockaddr*) &sa, sizeof(sa)) < 0) throw couldNotOpen();
  if (listen(fd_listen, 1) < 0) throw couldNotOpen();
  
  FD_ZERO(&fs);
  FD_SET(fd_listen, &fs);
  select(FD_SETSIZE, &fs, 0, 0, 0);
  
  if (sigIntOccured) 
    {
    sigIntOccured = false;
    close(fd_listen);
    throw userBreak();
    }
  
  fd = accept(fd_listen, (sockaddr*) &peer, &l);
  if (fd < 0) throw couldNotOpen();
  
  close(fd_listen);
  
  if (verbose) 
    printf("NET: Connected.\n");
  
  closeHandle = true;  
}

network::~network()
{
  if (closeHandle && (fd >= 0)) close(fd);
  if (buf) delete [] buf;
  
  if (localIV) delete [] localIV;
  if (remoteIV) delete [] remoteIV;
  
  if (localSecure) delete [] localSecure;
  if (remoteSecure) delete [] remoteSecure;
  if (localCryptor) end_mcrypt_cfb(localCryptor);
  if (remoteCryptor) end_mcrypt_cfb(remoteCryptor);
}

void network::read()
{
  int r, l, tmp;
  char *d;
  
  if (!handshook) handshake();
  
  r = ::read(fd, buf, BUF_SIZE);
  if( r< 0) throw couldNotRead();
  
  // Got a empty package
  if( r == 0 )
    {
      len = 0;
      return;
    }
  
  // Get the package size
  memcpy(&tmp, buf, sizeof(int));
  len = ntohl(tmp);
  d = buf + r;	
  l = len - (r - 4);
  
  if( len > BUF_SIZE  || len < 0 )
    {
      len = 0;
      return;
    }
  
  while(l > 0)
    {
      r = ::read(fd, d, l);
      if (r < 0) throw couldNotRead();
      l -= r; d += r;
    }
  
  if( encryption != -1 )
    mdecrypt_cfb(localCryptor, buf + 4, len);
}

int network::getBufLen()
{
  return len;
}

char *network::getBuf()
{
  return buf + 4;
}

void network::write(char *d, int s)
{
  unsigned int tmp;
  
  if (!handshook) handshake();
  
  if( encryption != -1 )
    mcrypt_cfb(remoteCryptor, d + 4, s);
  
  tmp = htonl(s);
  
  memcpy(d, &tmp, sizeof(int));
  
  if (::write(fd, d, s + 4 ) != s + 4)
    {
      perror("network::write()");
      throw couldNotWrite();
    }
}

network::operator int()
 {
   return fd;
 }

void network::setLocalSecure(char *s)
{
  if( !localSecure )
    localSecure = strcpy(new char[strlen(s)+1], s);
}

void network::setRemoteSecure(char *s)
{
  if( !remoteSecure )
    remoteSecure = strcpy(new char[strlen(s)+1], s);
}
 
void network::setEncryption(int e)
{
  if ((e < -1) || (e > 19) || (e == 5)) throw unknownEncryption();
  encryption = e;
}

void network::handshake()
{
  if (handshook) return;
  
  try
   {
     electMaster();
   }    
  catch (couldNotRead)  { throw handshakeFailure(); }    
  catch (couldNotWrite) { throw handshakeFailure(); }
  
  
  if (verbose) 
    printf("NET: Election was held, we are %s.\n", master ? "master" : "client");
  
  if( encryption != -1 )
    {
      try
	{
	  if (master)
	    {
	      if (readInteger() != encryption) throw encryptionInconsistent();
	      writeInteger(encryption);
	    }    
	  else
	    {
	      writeInteger(encryption);
	      if (readInteger() != encryption) throw encryptionInconsistent();
	    } 
	}
      catch (couldNotRead)  { throw encryptionInconsistent(); } 
      catch (couldNotWrite) { throw encryptionInconsistent(); } 
    }
  
  if (verbose) 
    {
      if( encryption != -1)
	printf("NET: Encryption is ok: %s\n", mcrypt_get_algorithms_name(encryption));
      else
	printf("NET: Encryption is DISABLED\n");
    }
  
  if( encryption != -1 )
    {
      try 
	{
	  if (master)
	    {
	      createLocalIV();
	      readRemoteIV();
	    }
	  else
	    {
	      readRemoteIV();
	      createLocalIV();
	    }
	}
      catch (couldNotRead) { throw handshakeFailure(); }
      catch (couldNotWrite) { throw handshakeFailure(); }
      
      if (verbose) 
	printf("NET: IVs exchanged.\n");
      
      localCryptor  = init_mcrypt_cfb(encryption, localSecure, strlen(localSecure), localIV);
      if (localCryptor == 0) throw cryptError();
      
      remoteCryptor = init_mcrypt_cfb(encryption, remoteSecure, strlen(remoteSecure), remoteIV);
      if (remoteCryptor == 0) throw cryptError();
      
      try
	{  
	  if (master)
	    {
	      authClient();
	      authServer();
	    }
	  else
	    {
	      authServer();
	      authClient();
	    }
	}
      catch (couldNotRead)  { throw authFailure(); }
      catch (couldNotWrite) { throw authFailure(); }
    }
  
  if (verbose) 
    printf("NET: Authorized.\n");
  
  handshook = true;
}

void network::electMaster()
{
  int r1, r2;
  
  do
    {
      r1 = rand();
      writeInteger(r1);
      r2 = readInteger();
    }
  while (r1 == r2);
  
  master = (r1 > r2) ? true : false;
}

void network::authServer()
{
  int v;
  v = readInteger();
  mdecrypt_cfb(localCryptor, &v, sizeof(v));
  mcrypt_cfb(remoteCryptor, &v, sizeof(v));
  writeInteger(v);
}

void network::authClient()
{
  int v, v2, v3;

  v2 = v = rand();
  mcrypt_cfb(remoteCryptor, &v2, sizeof(v2));
  writeInteger(v2);
  v3 = readInteger();
  mdecrypt_cfb(localCryptor, &v3, sizeof(v3));
  
  if (v != v3) throw authFailure();
}
 
void network::createLocalIV()
{
  char *buf = new char[mcrypt_get_block_size(encryption)];
  
  for (int i = 0; i < mcrypt_get_block_size(encryption); i++)
    {
      buf[i] = (char) rand();
      
      if (::write(fd, &buf[i], sizeof(char)) != sizeof(char))
	throw couldNotWrite();
    }    
  
  localIV = buf;
} 

void network::readRemoteIV()
{
  char *buf = new char[mcrypt_get_block_size(encryption)];
  
  for (int i = 0; i < mcrypt_get_block_size(encryption); i++)
    {
      if (::read(fd, &buf[i], sizeof(char)) != sizeof(char)) 
	throw couldNotRead();
    }     
  
  remoteIV = buf;
}


void network::writeInteger(int i)
{
  int tmp = htonl(i);
  
  if (::write(fd, &tmp, sizeof(int)) != sizeof(int)) throw couldNotWrite();
}

int network::readInteger()
{
  int tmp;
  
  if (::read(fd, &tmp, sizeof(int)) != sizeof(int)) throw couldNotRead();
  return( ntohl(tmp) );
}
 
